import os
import sys
import glob
import shutil
import setuptools.command.develop as develop

"""
This module contains some little workaround for the problem that
`python setup.py develop` (and thus `pip install -e .`) is incompatible
with a package layout where the modules are living on the top level.
However, that top-level-layout is needed for backward compatibility.
"""

current_dir = os.path.dirname(os.path.abspath(sys.modules.get(__name__).__file__))
parent_dir = os.path.dirname(current_dir)


def install_develop_mode(dist):
    """
    `python setup.py develop` assumes a directory structure where the package
    to be installed lives in a subdirectory. This function assumes that this command
    has already been run. Now it remains to
        - adapt the path inside the <package_name>.egg-link-text file (-> parent_dir)
        - adapt the path inside easy_install.pth (-> parent_dir)
        - move the directory <package_name>.egg-info to parent_dir

    For safety reasons the following directory structure is assumed:

    <arbitrary_dir_name>/              <- the .egg-link will point here
    │
    ├── <package_name>/                <- all the source lives here
    │   ├── .git/
    │   ├── setup.py
    │   ├── setup_develop_mode.py
    │   └── ...
    └── <package_name.egg-info>/       <- autogenerated by setuptools, moved here by this function

    Note: No other directories or files are allowed inside <arbitrary_dir_name>.

    :param dist: setuptools.dist object
    :return:
    """
    # noinspection PyArgumentList
    dev_obj = develop.develop(dist)
    dev_obj.ensure_finalized()
    old_egg_path = dev_obj.egg_path
    new_egg_path = os.path.dirname(old_egg_path)

    # ensure the required directory structure is met
    sibling_dirs = glob.glob(os.path.join(parent_dir, "*"))

    # remove the trivial case:
    sibling_dirs.remove(current_dir)

    # allow egg-info dir from previous install (will be overwritten anyway)
    parent_egg_info = os.path.join(parent_dir, os.path.split(dev_obj.dist.egg_info)[1])
    if parent_egg_info in sibling_dirs:
        sibling_dirs.remove(parent_egg_info)

    if len(sibling_dirs) > 0:
        # maybe it would be convenient to allow some exceptions here like, e.g. for config files
        msg = "The following files/directories contradict the assumed directory structure:\n{}"
        msg = msg.format("\n".join(sibling_dirs))
        raise FileExistsError(msg)

    if dev_obj.dry_run:
        return

    # this heavily inspired by setuptools.commands.develop.develop.install_for_development
    with open(dev_obj.egg_link, "w") as f:
        f.write(new_egg_path + "\n" + dev_obj.setup_path)

    # now move the egg-info directory one level higher
    egg_info_dirs = glob.glob(os.path.join(current_dir, "*egg-info"))
    assert len(egg_info_dirs) == 1
    source = egg_info_dirs[0]
    target = os.path.join(parent_dir, os.path.split(source)[1])
    assert target == parent_egg_info

    shutil.rmtree(target, ignore_errors=True)
    shutil.move(source, target)

    # this is inspired by pip._internals.req.req_uninstall.UninstallPathSet.from_dist
    easy_install_pth = os.path.join(os.path.dirname(dev_obj.egg_link), 'easy-install.pth')

    with open(easy_install_pth, "r") as f:
        all_paths = f.read()

    if old_egg_path in all_paths:
        all_paths = all_paths.replace(old_egg_path, new_egg_path)
        with open(easy_install_pth, "w") as f:
            f.write(all_paths)



# -*- coding: utf-8 -*-
# Owlready2
# Copyright (C) 2013-2019 Jean-Baptiste LAMY
# LIMICS (Laboratoire d'informatique médicale et d'ingénierie des connaissances en santé), UMR_S 1142
# University Paris 13, Sorbonne paris-Cité, Bobigny, France

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import owlready2
from owlready2.namespace import *
from owlready2.namespace import _cache_entity
from owlready2.entity    import *
from owlready2.entity    import _inherited_property_value_restrictions

class _EquivalentToList(CallbackList):
  __slots__ = ["_indirect"]
  def __init__(self, l, obj, callback):
    CallbackList.__init__(self, l, obj, callback)
    self._indirect = None
    
  def _build_indirect(self):
    n = self._obj.namespace
    self._indirect = list({
      n.ontology._to_python(o)
      for o in n.world._get_obj_triples_transitive_sym(self._obj.storid, owl_equivalentindividual)
      if o != self._obj.storid
    })
    
  def indirect(self):
    if self._indirect is None: self._build_indirect()
    return self._indirect
  
  def self_and_indirect_equivalent(self):
    yield self._obj
    if self._indirect is None: self._build_indirect()
    yield from self._indirect
    
  
    
class Thing(metaclass = ThingClass):
  namespace = owl
  
  def get_name(self): return self._name
  def set_name(self, name):
    self._name = name
    self.namespace.world._refactor(self.storid, "%s%s" % (self.namespace._base_iri, name))
  name = property(get_name, set_name)
  
  def get_iri(self):
    if self.storid < 0: return ""
    return "%s%s" % (self.namespace._base_iri, self._name)
  def set_iri(self, new_iri):
    splitted = new_iri.rsplit("#", 1)
    if len(splitted) == 2:
      self.namespace = self.namespace.ontology.get_namespace("%s#" % splitted[0])
    else:
      splitted = new_iri.rsplit("/", 1)
      self.namespace = self.namespace.ontology.get_namespace("%s/" % splitted[0])
    self._name = splitted[1]
    self.namespace.world._refactor(self.storid, new_iri)
  iri = property(get_iri, set_iri)
  
  def __new__(Class, name = None, namespace = None, **kargs):
    if name:
      if isinstance(name, Thing):
        namespace = name.namespace
        name      = name.name
      else:
        namespace = namespace or (CURRENT_NAMESPACES.get() and CURRENT_NAMESPACES.get()[-1]) or Class.namespace
      if LOADING or (name == ""):
        already_existing = None
      else:
        already_existing = namespace.world["%s%s" % (namespace._base_iri, name)]
        
      if not already_existing is None:
        if not isinstance(already_existing, Class):
          if isinstance(Class, FusionClass): Classes =  Class.__bases__
          else:                              Classes = (Class,)
          for C in Classes:
            if not C in already_existing.is_a:
              already_existing.is_a._append(C)
              if not LOADING:
                already_existing.namespace.ontology._add_obj_triple_spo(already_existing.storid, rdf_type, C.storid)
                
        bases = ThingClass._find_base_classes(already_existing.is_a)
        if len(bases) == 1:
          already_existing.__class__ = bases[0]
        else:
          already_existing.__class__ = FusionClass._get_fusion_class(bases)
          
        if not LOADING:
          for attr, value in kargs.items(): setattr(already_existing, attr, value)
          
        return already_existing
      
    return _cache_entity(object.__new__(Class))
  
  def __init__(self, name = None, namespace = None, **kargs):
    if   isinstance(name, int):
      is_new = True
      self.namespace = namespace or (CURRENT_NAMESPACES.get() and CURRENT_NAMESPACES.get()[-1]) or self.__class__.namespace
      iri = ""
      self._name = ""
    elif name:
      is_new = not "storid" in self.__dict__
      if is_new: self.namespace = namespace or (CURRENT_NAMESPACES.get() and CURRENT_NAMESPACES.get()[-1]) or self.__class__.namespace
      iri = "%s%s" % (self.namespace._base_iri, name)
      self._name = name
    else:
      is_new = True
      self.namespace = namespace or (CURRENT_NAMESPACES.get() and CURRENT_NAMESPACES.get()[-1]) or self.__class__.namespace
      iri = self.namespace.world._new_numbered_iri("%s%s" % (self.namespace._base_iri, self.generate_default_name()))
      self._name = iri[len(self.namespace._base_iri):]
    
    if is_new:
      self.__dict__["_equivalent_to"] = None
      if isinstance(name, int): self.storid = name or self.namespace.world.graph.new_blank_node()
      else:                     self.storid = self.namespace.world._abbreviate(iri)
      self.namespace.world._entities[self.storid] = self
      if isinstance(self.__class__, FusionClass):
        self.__dict__["is_a"] = CallbackList(self.__class__.__bases__, self, Thing._instance_is_a_changed)
      else:
        self.__dict__["is_a"] = CallbackList([self.__class__], self, Thing._instance_is_a_changed)
        
      if not LOADING:
        if self.storid > 0: self.namespace.ontology._add_obj_triple_spo(self.storid, rdf_type, owl_named_individual)
        for parent in self.is_a:
          self.namespace.ontology._add_obj_triple_spo(self.storid, rdf_type, parent.storid)
          
        for attr, value in kargs.items(): setattr(self, attr, value)
    
  def generate_default_name(self): return self.__class__.name.lower()
  
  def _get_is_instance_of(self):    return self.is_a
  def _set_is_instance_of(self, v): self.is_a = v
  is_instance_of = property(_get_is_instance_of, _set_is_instance_of)
  
  def _instance_is_a_changed(self, old):
    new = set(self.is_a)
    old = set(old)
    
    for base in old - new:
      if not LOADING: self.namespace.ontology._del_obj_triple_spo(self.storid, rdf_type, base.storid)
      if isinstance(base, Construct): base._set_ontology(None)
    bases = ThingClass._find_base_classes(self.is_a)
    if len(bases) == 1:
      self.__class__ = bases[0]
    elif bases:
      self.__class__ = FusionClass._get_fusion_class(bases)
    else:
      self.__class__ = Thing
      list.insert(self.is_a, 0, Thing)
      
    for base in new - old:
      if isinstance(base, Construct): base = base._set_ontology_copy_if_needed(self.namespace.ontology, self.is_a)
      if not LOADING: self.namespace.ontology._add_obj_triple_spo(self.storid, rdf_type, base.storid)
      
  #def __attrs__(self): # Not Python standard, but used by EditObj
  
  def get_equivalent_to(self):
    if self._equivalent_to is None:
      self._equivalent_to = _EquivalentToList(
        [self.namespace.world._to_python(o, default_to_none = True)
         for o in self.namespace.world._get_obj_triples_sp_o(self.storid, owl_equivalentindividual)
        ], self, Thing._instance_equivalent_to_changed)
    return self._equivalent_to
  
  def set_equivalent_to(self, value): self.equivalent_to.reinit(value)

  # Cannot use property because it name-clashes with the Class similar property
  #equivalent_to = property(get_equivalent_to, set_equivalent_to)
  
  def _instance_equivalent_to_changed(self, old):
    new = frozenset(self._equivalent_to)
    old = frozenset(old)
    
    for x in old - new:
      self.namespace.ontology._del_obj_triple_spo(self.storid, owl_equivalentindividual, x.storid)
      if isinstance(x, Construct): x._set_ontology(None)
      else: # Invalidate it
        if x.equivalent_to._indirect:
          for x2 in x.equivalent_to._indirect: x2._equivalent_to._indirect = None
          x._equivalent_to._indirect = None
          
    for x in new - old:
      self.namespace.ontology._add_obj_triple_spo(self.storid, owl_equivalentindividual, x.storid)
      if isinstance(x, Construct): x = x._set_ontology_copy_if_needed(self.namespace.ontology, self._equivalent_to)
      else: # Invalidate it
        if x.equivalent_to._indirect:
          for x2 in x.equivalent_to._indirect: x2._equivalent_to._indirect = None
          x._equivalent_to._indirect = None
          
    self._equivalent_to._indirect = None # Invalidate, because the addition / removal may add its own equivalent.
    
  def differents(self):
    for c, s, p, o in self.namespace.world._get_obj_triples_cspo_cspo(None, None, rdf_type, owl_alldifferent):
      onto = self.namespace.world.graph.context_2_user_context(c)
      list_bnode = self.namespace.world._get_obj_triple_sp_o(s, owl_distinctmembers)
      storids = set(storid for storid, dropit in onto._parse_list_as_rdf(list_bnode))
      if self.storid in storids: yield onto._parse_bnode(s)
      
      
  def __getattr__(self, attr):
    if attr.startswith("INDIRECT_"):
      Prop = self.namespace.world._props.get(attr[9:])
      if not Prop:
        if   attr == "INDIRECT_equivalent_to":
          eq = self.equivalent_to
          if eq._indirect is None: eq._build_indirect()
          return eq._indirect
        elif (attr == "INDIRECT_is_a") or (attr == "INDIRECT_is_instance_of"):
          #return list({ ancestor for parent in self.is_a if isinstance(parent, ThingClass) for ancestor in parent.ancestors() })
          return list({ ancestor for parent in self.is_a for ancestor in parent.ancestors(True, True) })
        else: raise AttributeError("'%s' property is not defined." % attr)
        
      if Prop.is_functional_for(self.__class__):
        return Prop._get_indirect_value_for_individual(self)
      else:
        return list(Prop._get_indirect_values_for_individual(self))
      
    elif attr.startswith("INVERSE_"):
      Prop = self.namespace.world._props.get(attr[8:])
      if Prop.inverse:
        return getattr(self, Prop.inverse.python_name)
      else:
        if issubclass_python(Prop, InverseFunctionalProperty): self.__dict__[attr] = r = Prop._get_inverse_value_for_individual (self)
        else:                                                  self.__dict__[attr] = r = Prop._get_inverse_values_for_individual(self)
        return r
      
    else:
      Prop = self.namespace.world._props.get(attr)
      if not Prop:
        if attr == "equivalent_to": return self.get_equivalent_to() # Needed
        raise AttributeError("'%s' property is not defined." % attr)
      if Prop.is_functional_for(self.__class__): self.__dict__[attr] = r = Prop._get_value_for_individual (self)
      else:                                      self.__dict__[attr] = r = Prop._get_values_for_individual(self)
      return r
    
  def __setattr__(self, attr, value):
    if attr in SPECIAL_ATTRS:
      if   attr == "is_a":          self.is_a.reinit(value)
      elif attr == "equivalent_to": self.set_equivalent_to(value) # Needed
      else:                         super().__setattr__(attr, value)
    else:
      Prop = self.namespace.world._props.get(attr)
      if Prop:
        if Prop.is_functional_for(self.__class__):
          if   Prop._owl_type == owl_object_property:
            old_value = self.__dict__.get(attr, None)
            #if Prop.inverse_property and (not old_value is None):
            #  old_value.__dict__.pop(Prop.inverse_property.python_name, None) # Remove => force reloading; XXX optimizable
            #  self.namespace.ontology._del_obj_triple_spo(old_value.storid, Prop.inverse_property.storid, self.storid) # Also remove inverse
            if not old_value is None:
              if Prop.inverse_property:
                old_value.__dict__.pop(Prop.inverse_property.python_name, None) # Remove => force reloading; XXX optimizable
                self.namespace.ontology._del_obj_triple_spo(old_value.storid, Prop.inverse_property.storid, self.storid) # Also remove inverse
              else:
                old_value.__dict__.pop("INVERSE_%s" % Prop.python_name, None) # Remove => force reloading; XXX optimizable
                
            super().__setattr__(attr, value)
            
            if value is None:
              self.namespace.ontology._del_obj_triple_spo(self.storid, Prop.storid, None)
            else:
              self.namespace.ontology._set_obj_triple_spo(self.storid, Prop.storid, value.storid)
              if Prop.inverse_property: value.__dict__.pop(Prop.inverse_property.python_name, None) # Remove => force reloading; XXX optimizable
              
          elif Prop._owl_type == owl_data_property:
            old_value = self.__dict__.get(attr, None)
            
            super().__setattr__(attr, value)
            
            if value is None:
              self.namespace.ontology._del_data_triple_spod(self.storid, Prop.storid, None, None)
            else:
              self.namespace.ontology._set_data_triple_spod(self.storid, Prop.storid, *self.namespace.ontology._to_rdf(value))
              
        else:
          if not isinstance(value, list):
            if isinstance(Prop, AnnotationPropertyClass):
              if value is None: value = []
              else:             value = [value]
            else:
              raise ValueError("Property '%s' is not functional, cannot assign directly (use .append() or assign a list)." % attr)
          getattr(self, attr).reinit(value)
          
      elif attr.startswith("INVERSE_"):
        Prop = self.namespace.world._props.get(attr[8:])
        if Prop:
          if Prop.inverse: setattr(self, Prop.inverse.python_name, value)
          else:
            if issubclass_python(Prop, InverseFunctionalProperty):
              old_value = self.__dict__.get(attr, None)
              if not old_value is None:
                if Prop.is_functional_for(old_value): setattr(old_value, Prop.python_name, None)
                else:                                 getattr(old_value, Prop.python_name).remove(self)
              if not value is None:
                if Prop.is_functional_for(value): setattr(value, Prop.python_name, self)
                else:                             getattr(value, Prop.python_name).append(self)
            else:
              if not isinstance(value, list): raise ValueError("Property '%s' is not inverse functional, cannot assign directly (use .append() or assign a list)." % attr)
              getattr(self, attr).reinit(value)
            
        else:
          raise ValueError("Property '%s' do not exist, cannot use its INVERSE." % attr[8:])
        
      else: super().__setattr__(attr, value)
        
  def _get_instance_possible_relations(self, ignore_domainless_properties = False):
    for Prop in self.namespace.world._reasoning_props.values():
      all_domains = set(Prop.domains_indirect())
      if ignore_domainless_properties and (not all_domains):
        for restrict in _inherited_property_value_restrictions(self, Prop, set()):
          yield Prop
          break
      else:
        for domain in all_domains:
          if not domain._satisfied_by(self): break
        else:
          yield Prop
          
  def get_properties(self):
    l = set()
    for storid in self.namespace.world._get_triples_s_p(self.storid):
      Prop = self.namespace.world._get_by_storid(storid)
      if not Prop is None: # None is is-a
        l.add(Prop)
    for storid in self.namespace.world._get_obj_triples_o_p(self.storid):
      Prop = self.namespace.world._get_by_storid(storid)
      if not Prop is None: # None is is-a
        if Prop._inverse_property:
          l.add(Prop._inverse_property)
    return l
  
  def INDIRECT_get_properties(self):
    l = self.get_properties()
    l.update(self.__class__.INDIRECT_get_class_properties())
    return l
  
  def __dir__(self):
    return set(object.__dir__(self)) | { Prop.python_name for Prop in self.get_properties() }
  
  def get_inverse_properties(self):
    for s,p,o in self.namespace.world._get_obj_triples_spo_spo(None, None, self.storid):
      Prop    = self.namespace.world._get_by_storid(p)
      if not Prop is None: # None is is-a
        subject = self.namespace.world._get_by_storid(s)
        yield subject, Prop
        


class Nothing(Thing): pass

class NamedIndividual(Thing): pass


class FusionClass(ThingClass):
  ontology = anonymous
  
  def __repr__(self): return "<FusionClass %s>" % ", ".join(repr(c) for c in self.__bases__)
  
  @staticmethod
  def _get_fusion_class(Classes0):
    Classes0 = tuple(sorted(Classes0, key = lambda Class: Class.__name__))
    Class = Classes0[0].namespace.world._fusion_class_cache.get(Classes0)
    if Class: return Class
    
    Classes = _keep_most_specific(Classes0, consider_equivalence = True)
    try:
      fusion_class = FusionClass._create_fusion_class(Classes0, Classes)
      
    except: # Too complex hierarchy => do not consider equivalent classes
      Classes = _keep_most_specific(Classes0, consider_equivalence = False)
      fusion_class = FusionClass._create_fusion_class(Classes0, Classes)
      
    return fusion_class
  
  @staticmethod
  def _create_fusion_class(Classes0, Classes):
    global _NEXT_FUSION_CLASS_ID
    
    if len(Classes) == 1:
      Class = tuple(Classes)[0]
      Class.namespace.world._fusion_class_cache[Classes0] = Class
      return Class
    
    Classes = tuple(sorted(Classes, key = lambda Class: Class.__name__))
    fusion_class = Classes0[0].namespace.world._fusion_class_cache.get(Classes)
    if fusion_class: return fusion_class
    #name = "_AND_".join(Class.__name__ for Class in Classes) # Causes name clashes if several world have Fusion Class with the same OWL classes and thus the same names
    name = "FusionClass%s" % _NEXT_FUSION_CLASS_ID
    _NEXT_FUSION_CLASS_ID += 1
    
    with anonymous: # Force triple insertion into anonymous
      try:
        fusion_class = FusionClass(name, Classes, { "namespace" : anonymous })
      except TypeError:
        fusion_class = ThingClass(name, Classes, { "namespace" : anonymous })
        
    Classes0[0].namespace.world._fusion_class_cache[Classes0] = Classes0[0].namespace.world._fusion_class_cache[Classes] = fusion_class
    return fusion_class
  
_NEXT_FUSION_CLASS_ID = 1



# -*- coding: utf-8 -*-
# Owlready2
# Copyright (C) 2013-2019 Jean-Baptiste LAMY
# LIMICS (Laboratoire d'informatique médicale et d'ingénierie des connaissances en santé), UMR_S 1142
# University Paris 13, Sorbonne paris-Cité, Bobigny, France

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from owlready2.namespace import *
from owlready2.prop      import *
from owlready2.prop      import _CLASS_PROPS

# import weakref
# _annot_axiom_cache = weakref.WeakValueDictionary()

# class _AnnotAxiom(object):
#   def __init__(self, source, property, target, target_d):
#     self.namespace = source.namespace.ontology
#     self.source    = source
#     self.property  = property
#     self.target    = target
#     self.target_d  = target_d
#     self.bnode     = self.search_bnode()
    
#   def search_bnode(self):
#     if self.target_d is None:
#       for bnode in self.namespace._get_obj_triples_po_s(rdf_type, owl_axiom):
#         for p, o in self._get_obj_triples_s_po(bnode):
#           if   (p == owl_annotatedsource)   and (o != self.source):   break
#           elif (p == owl_annotatedproperty) and (o != self.property): break
#           elif (p == owl_annotatedtarget)   and (o != self.target):   break
#       else:
#         return bnode
#     else:
#       for bnode in self.namespace._get_obj_triples_po_s(rdf_type, owl_axiom):
#         for p, o, d in self._get_triples_s_pod(bnode):
#           if   (p == owl_annotatedsource)   and (o != self.source):   break
#           elif (p == owl_annotatedproperty) and (o != self.property): break
#           elif (p == owl_annotatedtarget)   and (o != self.target):   break
#       else:
#         return bnode

class _AnnotList(CallbackListWithLanguage):
  __slots__ = ["namespace", "_property", "_target", "_target_d", "_annot", "_od_2_bnode"]
  def __init__(self, l, source, property, target, target_d, annot, namespace, od_2_bnode):
    list.__init__(self, l)
    self.namespace   = namespace
    self._obj        = source
    self._property   = property
    self._target     = target
    self._target_d   = target_d
    self._annot      = annot
    self._od_2_bnode = od_2_bnode
    
  def _callback(self, obj, old):
    old = set(old)
    new = set(self)
    
    if isinstance(obj, _AnnotList): # Annotate an annotation
      storid = obj._od_2_bnode[self._target, self._target_d]
    else:
      storid = obj.storid
    
    # Add before, in order to avoid destroying the axiom and then recreating, if all annotations are modified
    for added in new - old:
      o, d = obj.namespace.ontology._to_rdf(added)
      bnode = obj.namespace.ontology._add_annotation_axiom(storid, self._property, self._target, self._target_d, self._annot, o, d)
      self._od_2_bnode[o, d] = bnode
      
    for removed in old - new:
      obj.namespace.ontology._del_annotation_axiom(storid, self._property, self._target, self._target_d, self._annot, *obj.namespace.ontology._to_rdf(removed))
    
    
class AnnotationPropertyClass(PropertyClass):
  _owl_type = owl_annotation_property
  inverse_property = inverse = None
  
  def __getitem__(Annot, entity):
    if isinstance(entity, tuple):
      source, property, target = entity
      
      if   hasattr(source, "storid"):
        namespace = source.namespace # if Annot is in owl_world (e.g. comment), use the namespace of the source
      elif isinstance(source, _AnnotList):
        namespace = source._obj.namespace
      else:
        namespace = Annot.namespace
        
      if hasattr(property, "storid"): property = property.storid
      target, target_d = namespace.world._to_rdf(target)
      
      if   hasattr(source, "storid"):
        source_orig = source
        source      = source.storid
      elif isinstance(source, _AnnotList):
        source_orig = source
        source      = source._od_2_bnode[target, target_d]
        
      l = []
      
      od_2_bnode = {}
      for bnode in namespace.world._get_annotation_axioms(source, property, target, target_d):
        for o, d in namespace.world._get_triples_sp_od(bnode, Annot.storid):
          l.append(namespace.world._to_python(o, d))
          od_2_bnode[o, d] = bnode
          
      return _AnnotList(l, source_orig, property, target, target_d, Annot.storid, namespace, od_2_bnode)
    
    else:
      if Annot is entity.namespace.world._props.get(Annot._python_name) and not isinstance(entity, Construct): # use cached value
        r = getattr(entity, Annot._python_name)
        if isinstance(r, list): return r # May not be a list if hacked (e.g. Concept.terminology)
      return Annot._get_values_for_individual(entity)
    
  def __setitem__(Annot, index, values):
    if not isinstance(values, list): values = [values]
    Annot[index].reinit(values)
    
  def __call__(Prop, type, c, *args):
    raise ValueError("Cannot create a property value restriction on an annotation property!")
      
  def _get_indirect_values_for_individual(Prop, entity):
    values = [entity.namespace.ontology._to_python(o, d)
              for P in Prop.descendants(world = entity.namespace.world)
              for o, d in entity.namespace.world._get_triples_sp_od(entity.storid, P.storid)]
    return values
  
  _get_indirect_values_for_class = _get_indirect_values_for_individual
  

class AnnotationProperty(Property, metaclass = AnnotationPropertyClass):
  namespace = owl
  
  @classmethod
  def is_functional_for(Prop, o): return False


_CLASS_PROPS.add(AnnotationProperty)

  
class comment               (AnnotationProperty): namespace = rdfs
class label                 (AnnotationProperty): namespace = rdfs
class backwardCompatibleWith(AnnotationProperty): namespace = owl
class deprecated            (AnnotationProperty): namespace = owl
class incompatibleWith      (AnnotationProperty): namespace = owl
class isDefinedBy           (AnnotationProperty): namespace = rdfs
class priorVersion          (AnnotationProperty): namespace = owl
class seeAlso               (AnnotationProperty): namespace = rdfs
class versionInfo           (AnnotationProperty): namespace = owl




# -*- coding: utf-8 -*-
# Owlready
# Copyright (C) 2013-2019 Jean-Baptiste LAMY
# LIMICS (Laboratoire d'informatique médicale et d'ingénierie des connaissances en santé), UMR_S 1142
# University Paris 13, Sorbonne paris-Cité, Bobigny, France

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

#cp -f ./owlready2/test/test.owl /tmp/t.owl
#cp -f ./owlready2/test/test_inverse.owl /tmp/t.owl
#cp -f ./owlready2/test/test_reasoning.owl /tmp/t.owl
#cp -f ./owlready2/test/test_breakline.owl /tmp/t.owl
#cp -f ~/telechargements/base_med/obi.owl /tmp/t.owl

#rapper /tmp/t.owl > /tmp/rap.nt   ; python -c 'b = open("/tmp/rap.nt", "rb").read(); lb = b.split(b"\n"); ls = [i.replace(b"\\" + b"\"", b"\\\\" + b"\"").decode("unicode-escape").replace("\n", "\\n") for i in lb]; s = "\n".join(ls); open("/tmp/rap2.nt", "w").write(s)'
#python ./owlready2/rdfxml_2_ntriples.py /tmp/t.owl > /tmp/py.nt
#python ./owlready2/ntriples_diff.py /tmp/rap.nt /tmp/py.nt --short

"""ntriples_diff.py

Compute the differences between 2 NTriples files. Take care of symmetric relations
and different blank node naming.

Usage:

  ntriple_diff.py onto.nt : canonize the NTriples file.

  ntriple_diff.py onto1.nt onto2.nt : compare the two NTriples files.

Options:

  --short : shortify IRI and literal (more Human readable).
"""

import sys, os, warnings


def shortify(triples):
  triples2 = []
  for s,p,o, l in triples:
    if "#" in s: s = "<%s" % s.split("#", 1)[-1]
    if "#" in p: p = "<%s" % p.split("#", 1)[-1]
    if o.startswith('"'):
      o = o[1:].rsplit('"', 1)[0]
    elif "#" in o:
      o = "<%s" % o.split("#", 1)[-1]
    triples2.append((s,p,o, l))
  return triples2


symmetric_predicates = {
  "<http://www.w3.org/2002/07/owl#complementOf>",
  "<http://www.w3.org/2002/07/owl#inverseOf>",
  "<http://www.w3.org/2002/07/owl#equivalentClass>",
  "<http://www.w3.org/2002/07/owl#equivalentProperty>",
  "<http://www.w3.org/2002/07/owl#sameAs>",
  "<http://www.w3.org/2002/07/owl#disjointWith>",
  }

useless_triples = {
  ("<http://www.w3.org/1999/02/22-rdf-syntax-ns#PlainLiteral>", "<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>", "<http://www.w3.org/2000/01/rdf-schema#Datatype>"),
  ("<http://www.w3.org/2000/01/rdf-schema#Literal>", "<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>", "<http://www.w3.org/2000/01/rdf-schema#Datatype>"),
  ("<http://www.w3.org/2001/XMLSchema#float>", "<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>", "<http://www.w3.org/2000/01/rdf-schema#Datatype>"),
  ("<http://www.w3.org/2001/XMLSchema#integer>", "<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>", "<http://www.w3.org/2000/01/rdf-schema#Datatype>"),
  ("<http://www.w3.org/2001/XMLSchema#nonNegativeInteger>", "<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>", "<http://www.w3.org/2000/01/rdf-schema#Datatype>"),
  ("<http://www.w3.org/2001/XMLSchema#string>", "<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>", "<http://www.w3.org/2000/01/rdf-schema#Datatype>"),

  }

class Blank(object):
  def __init__(self):
    self.triples = set()
    self.name    = None
    
  def add_triple(self, s,p,o):
    if isinstance(s, Blank): s = "_"
    if isinstance(o, Blank): o = "_"
    self.triples.add((s,p,o))
    
  def get_name(self):
    if not self.name: self.name = "_:%s" % hash(frozenset(self.triples))
    return self.name
  
  
def canonize(nt):
  triples = []
  blanks  = {}
  l       = 0
  nt = nt.replace("/>", ">") # Owlready remove trailing /
  for line in nt.split("\n"):
    if not line: continue
    l += 1
    #print(l, line)
    if line.startswith("#") or not line: continue
    line = line[:-1].strip()
    s,p,o = line.split(None, 2)
    if o.endswith("^^<http://www.w3.org/1999/02/22-rdf-syntax-ns#PlainLiteral>"): o = o.rsplit("^^", 1)[0]
    if s.startswith("_"): s = blanks.get(s) or Blank()
    if o.startswith("_"): o = blanks.get(o) or Blank()
    if isinstance(s, Blank): s.add_triple(s,p,o)
    if isinstance(o, Blank): o.add_triple(s,p,o)
    triples.append((s,p,o, l))
    if p in symmetric_predicates:
      if isinstance(s, Blank): s.add_triple(o,p,s)
      if isinstance(o, Blank): o.add_triple(o,p,s)
      triples.append((o,p,s, 0))
      
      
  triples2 = []
  for s,p,o, l in triples:
    if isinstance(s, Blank): s = s.get_name()
    if isinstance(o, Blank): o = o.get_name()
    triples2.append((s,p,o, l))
    
  #triples2.sort()
  return triples2
  

def diff(nt1, nt2):
  triples1 = canonize(nt1)
  triples2 = canonize(nt2)
  
  triples1_set = set((s,p,o) for (s,p,o, l) in triples1)
  triples2_set = set((s,p,o) for (s,p,o, l) in triples2)
  
  removed = []
  added   = []
  
  for s,p,o, l in triples1:
    if (s,p,o) in useless_triples: continue
    if not (s,p,o) in triples2_set:
      removed.append((s,p,o, l))
      
  for s,p,o, l in triples2:
    if (s,p,o) in useless_triples: continue
    if not (s,p,o) in triples1_set:
      added.append((s,p,o, l))
      
  return removed, added


if __name__ == "__main__":
  
  if "--short" in sys.argv:
    sys.argv.remove("--short")
    short = True
  else:
    short = False
    
  if   len(sys.argv) == 2:
    nt = open(sys.argv[1]).read()
    triples = canonize(nt)
    
    if short: triples = shortify(triples)
    
    for s,p,o in sorted(triples):
      print(s,p,o, ".")
      
  elif len(sys.argv) == 3:
    nt1 = open(sys.argv[1]).read()
    nt2 = open(sys.argv[2]).read()
    removed, added = diff(nt1, nt2)
    
    if short:
      removed = shortify(removed)
      added   = shortify(added)
      
    for s,p,o, l in removed:
      if l: print("-", s, p, o, ". # line", l)
    print()
    for s,p,o, l in added:
      if l: print("+", s, p, o, ". # line", l)
    



# -*- coding: utf-8 -*-
# Owlready2
# Copyright (C) 2013-2019 Jean-Baptiste LAMY
# LIMICS (Laboratoire d'informatique médicale et d'ingénierie des connaissances en santé), UMR_S 1142
# University Paris 13, Sorbonne paris-Cité, Bobigny, France

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import owlready2
from owlready2.base      import *
from owlready2.namespace import *
from owlready2.namespace import _cache_entity



class _EquivalentToList(CallbackList):
  __slots__ = ["_indirect"]
  def __init__(self, l, obj, callback):
    CallbackList.__init__(self, l, obj, callback)
    self._obj      = obj
    self._indirect = None
    
  def _build_indirect(self):
    n = self._obj.namespace
    self._indirect = list(set( n.ontology._to_python(o, main_type = self._obj.__class__)
                               for o in n.world._get_obj_triples_transitive_sym(self._obj.storid, self._obj._owl_equivalent)
                               if o != self._obj.storid ))
    
  def indirect(self):
    if self._indirect is None: self._build_indirect()
    return self._indirect
  
  def self_and_indirect_equivalent(self):
    yield self._obj
    if self._indirect is None: self._build_indirect()
    yield from self._indirect
    

class _DisjointUnionList(CallbackList):
  __slots__ = ["_bn"]
  def __init__(self, l, obj, bn):
    CallbackList.__init__(self, l, obj, self.callback)
    self._bn  = bn
    self._obj = obj
    
  def callback(self, obj, old):
    obj.namespace.ontology._del_list(self._bn)
    obj.namespace.ontology._set_list(self._bn, self)
    
  

class EntityClass(type):
  namespace = owlready
  
  def get_name(Class): return Class._name
  def set_name(Class, name):
    Class._name = name
    Class.namespace.world._refactor(Class.storid, "%s%s" % (Class.namespace._base_iri, name))
  name = property(get_name, set_name)
  
  def get_iri(Class): return "%s%s" % (Class.namespace._base_iri, Class._name)
  def set_iri(Class, new_iri):
    splitted = new_iri.rsplit("#", 1)
    if len(splitted) == 2:
      Class.namespace = Class.namespace.ontology.get_namespace("%s#" % splitted[0])
    else:
      splitted = new_iri.rsplit("/", 1)
      Class.namespace = Class.namespace.ontology.get_namespace("%s/" % splitted[0])
    Class._name = splitted[1]
    Class.namespace.world._refactor(Class.storid, new_iri)
  iri = property(get_iri, set_iri)
  
  
  _owl_type         = owl_class
  _rdfs_is_a        = rdfs_subclassof
  _owl_equivalent   = owl_equivalentclass
  _owl_disjointwith = owl_disjointwith
  _owl_alldisjoint  = owl_alldisjointclasses
  
  @staticmethod
  def _find_base_classes(is_a):
    bases = tuple(Class for Class in is_a if not isinstance(Class, Construct))
    if len(bases) > 1:
      # Must use sorted() and not sort(), because bases is accessed during the sort
      return tuple(sorted(bases, key = lambda Class: sum(issubclass_python(Other, Class) for Other in bases)))
    return bases
  
  def mro(Class):
    try: return type.mro(Class)
    except TypeError:
      mro = [Class]
      for base in Class.__bases__:
        for base_mro in base.__mro__:
          if base_mro in mro: mro.remove(base_mro)
          mro.append(base_mro)
      return mro
    
  def __new__(MetaClass, name, superclasses, obj_dict):
    namespace = obj_dict.get("namespace") or (CURRENT_NAMESPACES.get() and CURRENT_NAMESPACES.get()[-1]) or superclasses[0].namespace
    storid    = obj_dict.get("storid")    or namespace.world._abbreviate("%s%s" % (namespace._base_iri, name))
    
    if "is_a" in obj_dict:
      _is_a = [*superclasses, *obj_dict["is_a"]]
      superclasses = MetaClass._find_base_classes(_is_a) or (Thing,)
    else:
      if len(superclasses) > 1:
        _is_a = superclasses = MetaClass._find_base_classes(superclasses) or (Thing,)
      else:
        _is_a = superclasses

    if LOADING:
      Class = namespace.world._entities.get (storid)
    else:
      Class = namespace.world._get_by_storid(storid)
      
    equivalent_to = obj_dict.pop("equivalent_to", None)
    
    if Class is None:
      if not LOADING:
        for base in _is_a:
          if isinstance(base, Construct):
            base = base._set_ontology_copy_if_needed(namespace.ontology, _is_a)
        
      _is_a = CallbackList(_is_a, None, MetaClass._class_is_a_changed)
      obj_dict.update(
        _name          = name,
        namespace      = namespace,
        storid         = storid,
        is_a           = _is_a,
        _equivalent_to = None,
      )
      
      Class = namespace.world._entities[storid] = _is_a._obj = type.__new__(MetaClass, name, superclasses, obj_dict)
      _cache_entity(Class)
      
      if not LOADING:
        namespace.ontology._add_obj_triple_spo(storid, rdf_type, MetaClass._owl_type)
        for parent in _is_a: Class._add_is_a_triple(parent)
        
    else:
      if not MetaClass is Class.__class__: Class.__class__ = MetaClass
      if (Class.is_a != _is_a) and (_is_a != (Thing,)):
        Class.is_a.extend([i for i in _is_a if not i in Class.is_a])
        
    if equivalent_to:
      if isinstance(equivalent_to, list): Class.equivalent_to.extend(equivalent_to)
      
    return Class
  
  def _add_is_a_triple(Class, base):
    Class.namespace.ontology._add_obj_triple_spo(Class.storid, Class._rdfs_is_a, base.storid)
    
  def _del_is_a_triple(Class, base):
    Class.namespace.ontology._del_obj_triple_spo(Class.storid, Class._rdfs_is_a, base.storid)
    
  def __init__(Class, name, bases, obj_dict):
    if "defined_class" in obj_dict:
      Class.defined_class = obj_dict["defined_class"] # Before the other!
    for k, v in obj_dict.items():
      if k in SPECIAL_ATTRS: continue
      Prop = Class.namespace.world._props.get(k)
      if Prop is None:
        type.__setattr__(Class, k, v)
      else:
        try: delattr(Class, k) # Remove the value initially stored by obj_dict in __new__
        except: pass
        setattr(Class, k, v)
        
  def get_equivalent_to(Class):
    if Class._equivalent_to is None:
      Class._equivalent_to = _EquivalentToList(
          [Class.namespace.world._to_python(o, main_type = Class.__class__, default_to_none = True)
           for o in Class.namespace.world._get_obj_triples_sp_o(Class.storid, Class._owl_equivalent)
          ], Class, Class.__class__._class_equivalent_to_changed)
    return Class._equivalent_to
  
  def set_equivalent_to(Class, value): Class.equivalent_to.reinit(value)
  
  equivalent_to = property(get_equivalent_to, set_equivalent_to)
  
  def get_indirect_equivalent_to(Class):
    eq = Class.equivalent_to
    if eq._indirect is None: eq._build_indirect()
    return eq._indirect
  INDIRECT_equivalent_to = property(get_indirect_equivalent_to)
  
  def _class_equivalent_to_changed(Class, old):
    for Subclass in Class.descendants(True, True):
      _FUNCTIONAL_FOR_CACHE.pop(Subclass, None)
      
    new = frozenset(Class._equivalent_to)
    old = frozenset(old)
    
    for x in old - new:
      Class.namespace.ontology._del_obj_triple_spo(Class.storid, Class._owl_equivalent, x.storid)
      if isinstance(x, Construct): x._set_ontology(None)
      else: # Invalidate it
        if not x.equivalent_to._indirect is None:
          for x2 in x.equivalent_to._indirect:
            if not isinstance(x2, Construct): x2._equivalent_to._indirect = None
          x._equivalent_to._indirect = None
      
    for x in new - old:
      if isinstance(x, Construct):
        x = x._set_ontology_copy_if_needed(Class.namespace.ontology, Class._equivalent_to)
          
      else: # Invalidate it
        if not x.equivalent_to._indirect is None:
          for x2 in x.equivalent_to._indirect:
            if not isinstance(x2, Construct): x2._equivalent_to._indirect = None
          x._equivalent_to._indirect = None
      Class.namespace.ontology._add_obj_triple_spo(Class.storid, Class._owl_equivalent, x.storid)
      
    Class._equivalent_to._indirect = None # Invalidate, because the addition / removal may add its own equivalent.
    
  def __setattr__(Class, attr, value):
    if attr == "is_a":
      old = Class.is_a
      type.__setattr__(Class, "is_a", CallbackList(value, Class, Class.__class__._class_is_a_changed))
      Class._class_is_a_changed(old)
      return

    type.__setattr__(Class, attr, value)
    
  def _class_is_a_changed(Class, old):
    for Subclass in Class.descendants(True, True):
      _FUNCTIONAL_FOR_CACHE.pop(Subclass, None)
      
    new = frozenset(Class.is_a)
    old = frozenset(old)
    for base in old - new:
      if not LOADING: Class._del_is_a_triple(base)
      if isinstance(base, Construct): base._set_ontology(None)
      
    bases = Class._find_base_classes(Class.is_a)
    if bases:
      Class.__bases__ = bases
    else:
      if   isinstance(Class, ThingClass):
        if not Class is Thing:
          Class.__bases__ = (Thing,)
          list.insert(Class.is_a, 0, Thing)
      elif isinstance(Class, ObjectPropertyClass):
        Class.__bases__ = (ObjectProperty,)
        list.insert(Class.is_a, 0, ObjectProperty)
      else:
        Class.__bases__ = (DataProperty,)
        list.insert(Class.is_a, 0, DataProperty)
        
    for base in new - old:
      if isinstance(base, Construct):
        if Class is Thing:
          base = base._set_ontology_copy_if_needed(CURRENT_NAMESPACES.get()[-1].ontology, Class.is_a)
        else:
          base = base._set_ontology_copy_if_needed(Class.namespace.ontology, Class.is_a)
      if not LOADING: Class._add_is_a_triple(base)
      
  def disjoints(Class):
    for c, s, p, o in Class.namespace.world._get_obj_triples_cspo_cspo(None, None, rdf_type, Class._owl_alldisjoint):
      onto = Class.namespace.world.graph.context_2_user_context(c)
      list_bnode = Class.namespace.world._get_obj_triple_sp_o(s, owl_members)
      storids = set(storid for storid, dropit in onto._parse_list_as_rdf(list_bnode))
      if Class.storid in storids: yield onto._parse_bnode(s)
      
    for c, s, p, o in Class.namespace.world._get_obj_triples_cspo_cspo(None, Class.storid, Class._owl_disjointwith, None):
      with LOADING: a = AllDisjoint((s, p, o), Class.namespace.world.graph.context_2_user_context(c), None)
      yield a # Must yield outside the with statement
      
    for c, s, p, o in Class.namespace.world._get_obj_triples_cspo_cspo(None, None, Class._owl_disjointwith, Class.storid):
      with LOADING: a = AllDisjoint((s, p, o), Class.namespace.world.graph.context_2_user_context(c), None)
      yield a
      
  def ancestors(Class, include_self = True, include_constructs = False):
    s = set()
    Class._fill_ancestors(s, include_self, include_constructs)
    return s
  
  def descendants(Class, include_self = True, only_loaded = False, world = None):
    s = set()
    if Class.namespace.world is owl_world:
      if world is None:
        import owlready2
        world = owlready2.default_world
      Class._fill_descendants(s, include_self, only_loaded, world, None)
    else:
      Class._fill_descendants(s, include_self, only_loaded, Class.namespace.world, Class.namespace.ontology)
    return s
  
  # def descendants2(Class, include_self = True, only_loaded = False, world = None):
  #   if (Class.namespace.world is owl_world) and (world is None):
  #     import owlready2
  #     world = owlready2.default_world
  #   else:
  #     world = Class.namespace.world

  #   if only_loaded:
  #     if include_self:
  #       return set(world.prepare_sparql("""SELECT ?c { ?c (rdfs:subClassOf|owl:equivalentClass|^owl:equivalentClass)* ?? . FILTER(ISIRI(?c)) . FILTER(LOADED(?c)) }""").execute_flat((Class,)))
  #     else:
  #       return set(world.prepare_sparql("""SELECT ?c { ?c (rdfs:subClassOf|owl:equivalentClass|^owl:equivalentClass)*/rdfs:subClassOf ?? . FILTER(ISIRI(?c)) . FILTER(LOADED(?c)) }""").execute_flat((Class,)))
  #   else:
  #     if include_self:
  #       return set(world.prepare_sparql("""SELECT ?c { ?c (rdfs:subClassOf|owl:equivalentClass|^owl:equivalentClass)* ?? . FILTER(ISIRI(?c)) }""").execute_flat((Class,)))
  #     else:
  #       return set(world.prepare_sparql("""SELECT ?c { ?c (rdfs:subClassOf|owl:equivalentClass|^owl:equivalentClass)*/rdfs:subClassOf ?? . FILTER(ISIRI(?c)) }""").execute_flat((Class,)))
    
  def _fill_ancestors(Class, s, include_self, include_constructs):
    if include_self:
      if not Class in s:
        s.add(Class)
        for equivalent in Class.equivalent_to.indirect():
          if isinstance(equivalent, EntityClass):
            if not equivalent in s: equivalent._fill_ancestors(s, True, include_constructs)
          elif include_constructs:
            if not equivalent in s: s.add(equivalent)
    if include_constructs:
      for parent in Class.is_a:
        if isinstance(parent, EntityClass):
          if not parent in s:
            parent._fill_ancestors(s, True, True)
        elif include_constructs and (not parent is object):
          if not parent in s: s.add(parent)
    else:
      for parent in Class.__bases__:
        if isinstance(parent, EntityClass):
          if not parent in s:
            parent._fill_ancestors(s, True, False)
            
  def _fill_descendants(Class, s, include_self, only_loaded, world, onto):
    if include_self:
      s.add(Class)
      for equivalent in Class.equivalent_to.indirect():
        if isinstance(equivalent, Class.__class__) and not equivalent in s:
          equivalent._fill_descendants(s, True, only_loaded, world, onto)
    
    for x in world._get_obj_triples_transitive_po(Class._rdfs_is_a, Class.storid):
      if not x < 0:
        if only_loaded:
          descendant = world._entities.get(x)
          if descendant is None: continue
        else:
          descendant = world._get_by_storid(x, None, Class.__class__, onto)
        if (descendant is Class): continue
        if not descendant in s:
          s.add(descendant)
          for equivalent in descendant.equivalent_to.indirect():
            if isinstance(equivalent, Class.__class__) and not equivalent in s:
              equivalent._fill_descendants(s, True, only_loaded, world, onto)
              
  def subclasses(Class, only_loaded = False, world = None):
    if Class is Thing:
      if world is None:
        import owlready2
        world = owlready2.default_world
        
      for x, in world.graph.db.execute(
          """SELECT q1.s FROM objs q1 WHERE q1.p=? AND q1.o=? AND ((q1.s > 0))
    AND (SELECT 1 FROM objs q2 INDEXED BY index_objs_sp
    WHERE q2.s=q1.s AND q2.p=? AND ((q2.o > 0) AND q2.o != ?)) IS NULL""",
          (rdf_type, owl_class, rdfs_subclassof, owl_thing)):
        if only_loaded:
          subclass = world._entities.get(x)
          if not subclass is None: yield subclass
        else:
          yield world._get_by_storid(x, None, ThingClass)
      
    else:
      world = world or Class.namespace.world
      if only_loaded:
        for x in world._get_obj_triples_po_s(Class._rdfs_is_a, Class.storid):
          if not x < 0:
            subclass = world._entities.get(x)
            if not subclass is None: yield subclass
            
      else:
        for x in world._get_obj_triples_po_s(Class._rdfs_is_a, Class.storid):
          if not x < 0:
            yield world._get_by_storid(x, None, ThingClass, Class.namespace.ontology)
        
  def constructs(Class, Prop = None):
    def _top_bn(onto, s):
      for relation in [rdf_first, rdf_rest, owl_complementof, owl_unionof, owl_intersectionof, owl_onclass, SOME, ONLY]:
        s2 = Class.namespace.world._get_obj_triple_po_s(relation, s)
        if not s2 is None: return _top_bn(onto, s2)
      try:
        return onto._parse_bnode(s)
      except:
        return None
      
    if Prop: Prop = Prop.storid
    for c,s,p,o in Class.namespace.world._get_obj_triples_cspo_cspo(None, None, Prop, Class.storid):
      if s < 0:
        
        onto = Class.namespace.world.graph.context_2_user_context(c)
        construct = _top_bn(onto, s)
        if not construct is None:
          yield construct
          
def issubclass_owlready(Class, Parent_or_tuple):
  try:
    if issubclass_python(Class, Parent_or_tuple): return True
  except TypeError: return False
  
  if isinstance(Class, EntityClass):
    if not isinstance(Parent_or_tuple, tuple): Parent_or_tuple = (Parent_or_tuple,)
    parent_storids = { Parent.storid for Parent in Parent_or_tuple }
    
    Class_parents = set(Class.namespace.world._get_obj_triples_transitive_sp(Class.storid, Class._rdfs_is_a))
    Class_parents.add(Class.storid)
    if not parent_storids.isdisjoint(Class_parents): return True
    
    equivalent_storids = { Equivalent.storid for Parent in Parent_or_tuple for Equivalent in Parent.equivalent_to.indirect() }
    if not equivalent_storids.isdisjoint(Class_parents): return True
    
  return False

issubclass = issubclass_owlready

def isinstance_python(obj, class_or_tuple):
  if isinstance(class_or_tuple, tuple):
    for c in class_or_tuple:
      if type.__instancecheck__(c, obj): return True
  else:
    return type.__instancecheck__(class_or_tuple, obj)
  

class ThingClass(EntityClass):
  namespace = owlready
  
  def __instancecheck__(Class, instance):
    if not hasattr(instance, "storid"): return False
    if Class is Thing: return super().__instancecheck__(instance)
    for C in instance.is_a:
      if isinstance(C, EntityClass) and issubclass(C, Class): return True
    return False
  
  def _satisfied_by(Class, x):
    return (isinstance(x, EntityClass) and issubclass(x, Class)) or isinstance(x, Class)
  
  def _get_class_possible_relations(Class):
    for Prop in Class.namespace.world._reasoning_props.values():
      for domain in Prop.domains_indirect():
        if not domain._satisfied_by(Class): break
      else:
        yield Prop
        
  def get_disjoint_unions(Class):
    if not "_disjoint_unions" in Class.__dict__:
      l = []
      for o in Class.namespace.world._get_obj_triples_sp_o(Class.storid, owl_disjointunion):
        xs = [Class.namespace.world._get_by_storid(storid) for storid, dropit in Class.namespace.world._parse_list_as_rdf(o)]
        du = _DisjointUnionList(xs, Class, o)
        l.append(du)
      Class._disjoint_unions = CallbackList(l, Class, Class.__class__._disjoint_union_changed)
    return Class._disjoint_unions
  def set_disjoint_unions(Class, dus): Class.disjoint_unions.reinit(dus)
  def _disjoint_union_changed(Class, old):
    new = Class._disjoint_unions
    for removed in old:
      if not removed in new:
        Class.namespace.ontology._del_list(removed._bn)
        Class.namespace.ontology._del_obj_triple_spo(Class.storid, owl_disjointunion, removed._bn)
    new2 = []
    for added in new:
      if not added in old:
        if not isinstance(new, _DisjointUnionList):
          bn = Class.namespace.world.new_blank_node()
          Class.namespace.ontology._set_list(bn, added)
          added = _DisjointUnionList(added, Class, bn)
        Class.namespace.ontology._add_obj_triple_spo(Class.storid, owl_disjointunion, added._bn)
      new2.append(added)  
    new._set(new2)
  disjoint_unions = property(get_disjoint_unions, set_disjoint_unions)
  
  
  # def instances(Class, world = None):
  #   if Class.namespace.world is owl_world:
  #     import owlready2
  #     world = (world or owlready2.default_world).world
  #     if Class is Thing: return world.individuals()
  #     return world.search(type = Class)
  #   return Class.namespace.world.search(type = Class)
  
  def instances(Class, world = None):
    if Class.namespace.world is owl_world:
      import owlready2
      world = (world or owlready2.default_world).world
      if Class is Thing: return world.individuals()
    else:
      world = Class.namespace.world
      
    q = world.prepare_sparql("""SELECT DISTINCT ?i { ?i a/(rdfs:subClassOf|owl:equivalentClass|^owl:equivalentClass)* ?? . }""")
    if sqlite3.sqlite_version_info[1] < 36: # Some SQLite3 versions have a bug on recursive queries with multiple unions
      q.sql = """
WITH RECURSIVE prelim1_objs(s) AS (VALUES (?1) 
UNION 
SELECT q.s FROM objs q, prelim1_objs rec WHERE q.p IN (9,33) AND q.o=rec.s),

prelim2_objs(s) AS (VALUES (?1) 
UNION
SELECT q.o FROM objs q, prelim1_objs rec WHERE q.p=33 AND q.s=rec.s)

SELECT q1.s FROM objs q1 WHERE q1.p=6 AND (q1.o IN (SELECT s FROM prelim1_objs) OR q1.o IN (SELECT s FROM prelim2_objs) ) ;
"""
    return list(q.execute_flat((Class,)))
  
  def direct_instances(Class, world = None):
    if Class.namespace.world is owl_world:
      import owlready2
      world = world or owlready2.default_world
      return [world._get_by_storid(s, None, Thing) for s in world._get_obj_triples_po_s(rdf_type, Class.storid)]
    return [Class.namespace.world._get_by_storid(s, None, Thing) for s in Class.namespace.world._get_obj_triples_po_s(rdf_type, Class.storid)]
  
  def get_class_properties(Class):
    l = set()
    for r in _property_value_restrictions(Class, None):
      if   r.property._class_property_some and ((r.type == VALUE) or (r.type == SOME) or ((r.type == EXACTLY) and r.cardinality >= 1) or ((r.type == MIN) and r.cardinality >= 1)):
        l.add(r.property)
      elif r.property._class_property_only and  (r.type == ONLY):
        l.add(r.property)
        
    for storid in Class.namespace.world._get_triples_s_p(Class.storid):
      Prop = Class.namespace.world._get_by_storid(storid)
      if not Prop is None: # None is is-a,...
        l.add(Prop)
    return l
  
  def INDIRECT_get_class_properties(Class):
    l = set()
    for r in _inherited_properties_value_restrictions(Class, None, set()):
      if   r.property._class_property_some and ((r.type == VALUE) or (r.type == SOME) or ((r.type == EXACTLY) and r.cardinality >= 1) or ((r.type == MIN) and r.cardinality >= 1)):
        l.add(r.property)
      elif r.property._class_property_only and  (r.type == ONLY):
        l.add(r.property)
        
    for storid in Class.namespace.world._get_triples_s_p(Class.storid):
      Prop = Class.namespace.world._get_by_storid(storid)
      if not Prop is None: # None is is-a,...
        l.add(Prop)
    return l
  
  def __and__(a, b): return And([a, b])
  def __or__ (a, b): return Or ([a, b])
  def __invert__(a): return Not(a)
  
  def __rshift__(Domain, Range):
    import owlready2.prop
    owlready2.prop._NEXT_DOMAIN_RANGE.set((Domain, Range))
    if isinstance(Range, ThingClass) or isinstance(Range, Construct):
      return owlready2.prop.ObjectProperty
    else:
      return owlready2.prop.DataProperty
    
  def get_defined_class(Class):
    v = Class.__dict__.get("__defined_class", None)
    if v is None:
      v = Class.namespace.ontology._get_data_triple_sp_od(Class.storid, owlready_defined_class)
      if v: v = from_literal(*v)
      else: v = False
      type.__setattr__(Class, "__defined_class", v)
    return v
  def set_defined_class(Class, defined_class):
    type.__setattr__(Class, "__defined_class", defined_class)
    if defined_class:
      Class.namespace.ontology._set_data_triple_spod(Class.storid, owlready_defined_class, *to_literal(True))
    else:
      Class.namespace.ontology._del_data_triple_spod(Class.storid, owlready_defined_class, "true", None)
  defined_class = property(get_defined_class, set_defined_class)
  
  def __getattr__(Class, attr):
    if attr.startswith("INDIRECT_"):
      if attr == "INDIRECT_is_a": return Class.ancestors(True, True)
      Prop = Class.namespace.world._props.get(attr[9:])
      if not Prop: raise AttributeError("'%s' property is not defined." % attr)
      #if Prop.is_functional_for(Class): return Prop._get_indirect_value_for_class(Class)
      if issubclass(Prop, FunctionalProperty): return Prop._get_indirect_value_for_class(Class)
      else:                                    return Prop._get_indirect_values_for_class(Class)
      
    else:
      Prop = Class.namespace.world._props.get(attr)
      if not Prop: raise AttributeError("'%s' property is not defined." % attr)
      if issubclass_python(Prop, AnnotationProperty):
        attr = "__%s" % attr # Do NOT cache as such in __dict__, to avoid inheriting annotations
        values = Class.__dict__.get(attr)
        if values is None:
          values = Prop._get_values_for_class(Class)
          type.__setattr__(Class, attr, values)
        return values
      
      #if Prop.is_functional_for(Class): return Prop._get_value_for_class (Class)
      if issubclass(Prop, FunctionalProperty): return Prop._get_value_for_class (Class)
      else:                                    return Prop._get_values_for_class(Class)
      
      
  def __setattr__(Class, attr, value):
    if attr in SPECIAL_ATTRS:
      super().__setattr__(attr, value)
      return

    if Class.namespace.world is owl_world: Prop = CURRENT_NAMESPACES.get()[-1].world._props.get(attr)
    else:                                  Prop = Class.namespace.world._props.get(attr)
    if Prop is None: raise AttributeError("'%s' property is not defined." % attr)
    
    if   value is None:               value = []
    elif not isinstance(value, list): value = [value]
    
    if isinstance(Prop, ReasoningPropertyClass):
      Class._on_class_prop_changed(Prop, Prop._get_values_for_class(Class), value)
    else:
      attr = "__%s" % attr # Do NOT cache as such in __dict__, to avoid inheriting annotations
      values = Class.__dict__.get(attr)
      if values is None:
        values = Prop._get_values_for_class(Class)
        type.__setattr__(Class, attr, values)
      values.reinit(value)
    
        
  def inverse_restrictions(Class, Prop = None):
    for construct in Class.constructs():
      if   isinstance(construct, Restriction) and ((Prop is None) or (construct.property is Prop)):
        if   construct.value is Class:
          yield from construct.subclasses()
        elif isinstance(construct.value, Or) and (Class in construct.value.Classes):
            yield from construct.subclasses()
            
      elif isinstance(construct, And):
        for subconstruct in construct.Classes:
          if isinstance(subconstruct, Restriction) and ((Prop is None) or (subconstruct.property is Prop)):
            if   subconstruct.value is Class:
              yield from construct.subclasses()
            elif isinstance(subconstruct.value, Or) and (Class in subconstruct.value.Classes):
              yield from construct.subclasses()
              
  # Class properties

  def _get_defined_construct(Class):
    parents = { c for c in Class.is_a if isinstance(c, ThingClass) }
    for r in Class.equivalent_to:
      if isinstance(r, And) and (parents.issubset(r.Classes)): break
    else: r = None
    return parents, r
  
  def _on_class_prop_changed(Class, Prop, old, new):
    old     = set(old)
    new     = set(new)
    removed = old - new
    inverse = Prop.inverse_property
    
    if Class.defined_class:
      parents, r = Class._get_defined_construct()
      
      if Prop._class_property_some:
        if r:
          for v in removed:
            for i in r.Classes:
              if isinstance(i, Restriction) and (i.property is Prop) and ((i.type == SOME) or (i.type == VALUE)) and (i.value is v):
                r.Classes.remove(i)
                break
              
        added = new - old
        if added:
          if not r:
            r = And(parents)
            Class.equivalent_to.append(r)
          for v in added:
            if isinstance(v, ThingClass): r.Classes.append(Prop.some(v))
            else:                         r.Classes.append(Prop.value(v))
      
      if Prop._class_property_only:
        if r:
          for r2 in r.Classes:
            if isinstance(r2, Restriction) and (r2.type == ONLY) and (r2.property is Prop): break
          else: r2 = None
        else: r2 = None
        
        only_classes   = [v for v in new if isinstance(v, EntityClass) or isinstance(v, Construct)]
        only_instances = [v for v in new if not v in only_classes]
        
        if only_instances: only_classes.append(OneOf(only_instances))
        if not only_classes:
          if r2 in r.Classes: r.Classes.remove(r2)
        else:
          if not r:
            r = And(parents)
            Class.equivalent_to.append(r)
          if len(only_classes) == 1:
            if r2: r2.value = only_classes[0]
            else:
              r2 = Prop.only(only_classes[0])
              r.Classes.append(r2)
          else:
            if r2:
              if isinstance(r2.value, Or): r2.value.Classes = only_classes
              else:                        r2.value = Or(only_classes)
            else:
              r2 = Prop.only(Or(only_classes))
              r.Classes.append(r2)
        
    else:
      if Prop._class_property_some:
        if removed:
          for r in list(_property_value_restrictions(Class, Prop)):
            if ((r.type == SOME) or (r.type == VALUE)) and (r.value in removed) and (r in Class.is_a):
              Class.is_a.remove(r)
              if r.type == VALUE:
                if isinstance(Prop, ObjectPropertyClass):
                  for r2 in r.value.is_a:
                    if isinstance(r2, Restriction) and ((r2.property is inverse) or (isinstance(r2.property, Inverse) and (r2.property.property is Prop))) and (r2.type == SOME) and (r2.value is Class):
                      r.value.is_a.remove(r2)
                      break
                    
        for v in new - old:
          if isinstance(v, EntityClass) or isinstance(v, Construct):
            Class.is_a.append(Prop.some(v))
          else:
            Class.is_a.append(Prop.value(v))
            if isinstance(Prop, ObjectPropertyClass):
              v.is_a.append(Inverse(Prop).some(Class))
            
      if Prop._class_property_only:
        for r in _property_value_restrictions(Class, Prop):
          if (r.type == ONLY): break
        else: r = None
        
        only_classes   = [v for v in new if isinstance(v, EntityClass) or isinstance(v, Construct)]
        only_instances = [v for v in new if not v in only_classes]
        
        if only_instances: only_classes.append(OneOf(only_instances))
        if not only_classes:
          if r in Class.is_a: Class.is_a.remove(r)
        else:
          if len(only_classes) == 1:
            if r: r.value = only_classes[0]
            else:
              r = Prop.only(only_classes[0])
              Class.is_a.append(r)
          else:
            if r:
              if isinstance(r.value, Or): r.value.Classes = only_classes
              else:                       r.value = Or(only_classes)
            else:
              r = Prop.only(Or(only_classes))
              Class.is_a.append(r)
              
      if Prop._class_property_relation:
        if Prop._owl_type == owl_object_property:
          for v in removed:
            Class.namespace.ontology._del_obj_triple_spo(Class.storid, Prop.storid, v.storid)
            if inverse:
              Class.namespace.ontology._del_obj_triple_spo(v.storid, inverse.storid, Class.storid) # Also remove inverse
              #if isinstance(v, EntityClass): v.__dict__.pop("__%s" % inverse.python_name, None) # Remove => force reloading; XXX optimizable
              #else:                          v.__dict__.pop(inverse.python_name, None) # Remove => force reloading; XXX optimizable
              
          for v in new - old:
            Class.namespace.ontology._add_obj_triple_spo(Class.storid, Prop.storid, v.storid)
            #if inverse:
            #  if isinstance(v, EntityClass): v.__dict__.pop("__%s" % inverse.python_name, None) # Remove => force reloading; XXX optimizable
            #  else:                          v.__dict__.pop(inverse.python_name, None) # Remove => force reloading; XXX optimizable
            
        else: # Data prop
          for v in removed:
            Class.namespace.ontology._del_data_triple_spod(Class.storid, Prop.storid, Class.namespace.ontology._to_rdf(v)[0], None)
            
          for v in new - old:
            Class.namespace.ontology._add_data_triple_spod(Class.storid, Prop.storid, *Class.namespace.ontology._to_rdf(v))
          
          
            


def _inherited_property_value_restrictions(x, Prop, already):
  if   isinstance(x, Restriction):
    if (Prop is None) or (x.property is Prop): yield x
    
  elif isinstance(x, EntityClass) or isinstance(x, Thing):
    # Need two passes in order to favor restriction on the initial class rather than those on the ancestor classes
    parentss = (x.is_a, list(x.equivalent_to.indirect()))
    for parents in parentss:
      for parent in parents:
        if isinstance(parent, Restriction):
          if (Prop is None) or (parent.property is Prop): yield parent
        
    for parents in parentss:
      for parent in parents:
        if (not isinstance(parent, Restriction)) and (not parent in already):
          already.add(parent)
          yield from _inherited_property_value_restrictions(parent, Prop, already)
          
  elif isinstance(x, And):
    for x2 in x.Classes:
      yield from _inherited_property_value_restrictions(x2, Prop, already)
      



# -*- coding: utf-8 -*-
# Owlready2
# Copyright (C) 2013-2019 Jean-Baptiste LAMY
# LIMICS (Laboratoire d'informatique médicale et d'ingénierie des connaissances en santé), UMR_S 1142
# University Paris 13, Sorbonne paris-Cité, Bobigny, France

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["EditedInstances", "OntologyInstanceEditor"]
import locale

import owlready2, owlready2.editor
from owlready2 import *
from owlready2.editor import _get_class_one_of, _is_abstract_class, VALUES_LISTERS

import editobj3, editobj3.undoredo as undoredo, editobj3.introsp as introsp, editobj3.field as field, editobj3.editor as editor
from editobj3.observe import *


class EditedInstances(object):
  def __init__(self, ontology, Class):
    self.ontology  = ontology
    self.namespace = ontology
    self.Class     = Class
    self.name      = editobj3.TRANSLATOR(Class.name.replace("_", " "))
    
  def get_instances(self):
    l = self.ontology.search(type = self.Class)
    l.sort(key = repr)
    return l
  instances = property(get_instances)
  
  def remove_instance(self, instance):
    destroy_entity(instance)
    
  def __str__(self): return self.name
  
  def details(self): return "%s (%s)" % (self.name, len(self.instances))
  
  def addable_values(self):
    return [introsp.NewInstanceOf(Class)
            for Class in self.Class.descendants()
            if (not _get_class_one_of(Class)) and (not _is_abstract_class(Class))]
  
descr = introsp.description(EditedInstances)
descr.set_details(EditedInstances.details)
descr.def_attr("instances", label = "", reorder_method = None, addable_values = "addable_values")
descr.def_attr("ontology" , field.HiddenField)
descr.def_attr("namespace", field.HiddenField)
descr.def_attr("Class"    , field.HiddenField)
descr.def_attr("name"     , field.HiddenField)


class TabPaneRepartitor(editor.PaneRepartitor):
  def __init__(self, instance_editor, tab_edited_class):
    self.instance_editor  = instance_editor
    self.tab_edited_class = tab_edited_class
    self._cache           = {}
    
  def is_displayed_in_other_tab(self, attribute, o_Class):
    Prop = default_world._props.get(attribute.name)
    if Prop: 
      for Class in o_Class.mro():
        values_lister = VALUES_LISTERS.get((Prop, Class))
        if values_lister:
          other_tab_classes = list(self.instance_editor.edited_classes)
          other_tab_classes.remove(self.tab_edited_class)
          return values_lister.range_match_classes(other_tab_classes)
    return False
  
  def _compute(self, o, attribute, field_class = None):
    o_Class = o.__class__
    if field_class is None: field_class = attribute.field_class_for(o)
    
    displayed_in_another_tab = False
    Prop = default_world._props.get(attribute.name)
    if Prop:
      if isinstance(o, type): mro = o.mro()
      else:                   mro = o_Class.mro()
      for Class in mro:
        values_lister = VALUES_LISTERS.get((Prop, Class))
        if values_lister:
          other_tab_classes = list(self.instance_editor.edited_classes)
          other_tab_classes.remove(self.tab_edited_class)
          displayed_in_another_tab = values_lister.range_match_classes(other_tab_classes)
          break
          
    displayed_in_hierarchy_pane = field_class.display_in_hierarchy_pane and not(displayed_in_another_tab)
    displayed_in_attribute_pane = field_class.display_in_attribute_pane or      displayed_in_another_tab
    r = self._cache[o_Class, attribute.name] = displayed_in_hierarchy_pane, displayed_in_attribute_pane
    return r
  
  def is_displayed_in_hierarchy_pane(self, attribute, o, field_class = None):
    cached = self._cache.get((o.__class__, attribute.name))
    if cached: return cached[0]
    return self._compute(o, attribute, field_class)[0]
  
  def is_displayed_in_attribute_pane(self, attribute, o, field_class = None):
    cached = self._cache.get((o.__class__, attribute.name))
    if cached: return cached[1]
    return self._compute(o, attribute, field_class)[1]


  
class OntologyInstanceEditor(editor.EditorTabbedDialog):
  _Qt_MODULE     = "owlready2.instance_editor_qt"
  _HTML_MODULE   = "owlready2.instance_editor_html"
  def __init__(self, gui = None, master = None, direction = "h", on_validate = None, edit_child_in_self = 1, undo_stack = None, on_close = None, menubar = True):
    super().__init__(gui, master, direction, on_validate, edit_child_in_self, undo_stack, on_close, menubar)
    self.set_default_size(1024, 752)
    self.add_to_menu(self.file_menu, 0, u"Save"      , self.on_save, accel = u"C-S", pos = 0)
    self.add_to_menu(self.file_menu, 0, u"Save as...", self.on_save_as, accel = u"C-S-S", pos = 1)
    self.add_separator_to_menu(self.file_menu, pos = 2)
    
    self.ontology          = None
    self.edited_classes    = []
    self.last_undoables    = []
    self.edited_instancess = {}
    
  def on_dialog_closed(self, *args):
    super().on_dialog_closed(*args)
    if not self.check_save(): return True
    return False
    
  def set_ontology(self, ontology, edited_classes = None):
    if not edited_classes: edited_classes = [
        Class for Class in ontology.classes()
        if not [parent for parent in Class.is_a if parent.namespace.ontology is ontology]]
    self.ontology          = ontology
    self.edited_classes    = []
    self.last_undoables    = []
    self.edited_instancess = {}
    # edit() must be called AFTER setting the pane repartitor, thus cannot be done by add_tab(),
    # and AFTER creating all tabs (since the pane repartitor needs the entire list of tabs)
    for edited_class in edited_classes: self.add_tab_for_class(edited_class)
    for edited_class in edited_classes: self.editor_panes[edited_class].edit(self.edited_instancess[edited_class])
    scan()
    
  def add_tab_for_class(self, Class):
    self.edited_classes.append(Class)
    edited_instance = self.edited_instancess[Class] = EditedInstances(self.ontology, Class)
    editor_pane = self.add_tab(Class, str(edited_instance))
    editor_pane.set_pane_repartitor(TabPaneRepartitor(self, Class))
    return editor_pane
    
  def on_save(self, *args):
    self.ontology.save()
    self.last_undoables = self.undo_stack.undoables[:]
    
  def on_save_as(self, *args):
    filename = self.prompt_save_filename()
    if not filename: return
    dirname, filename = os.path.split(filename)
    self.ontology.name = filename
    if onto_path[0] != dirname: onto_path.insert(0, dirname)
    self.on_save()

    



